
function gradient(prob::Problem{T},
                  comms::OptComms;
                 ) where {T}
    world_rank = comms.ranks.world
    compute_rank = comms.ranks.compute
    G = world_rank == 0 ? zeros(T, length(prob.state_vector)) : nothing
    finished_loop = false
    if world_rank == 0
        for (key, wrapper) in prob.equilibrium_wrappers
            @info "Computing the gradient at iteration $(prob.iteration)
            for $(equilibrium_type_string(wrapper)) state variables
            with method $(derivative_type_string(wrapper))"
            MPI.bcast(false, 0, comms.world_comm)
            MPI.bcast(wrapper, 0, comms.world_comm)
            var_indices = findall(v -> v.equilibrium_key == key, prob.state_vector)
            MPI.bcast(var_indices, 0, comms.world_comm)
            eq_targets = filter(p -> prob.derived_geometries[p.second.geometry_key].parent_key == key, prob.targets)
            MPI.bcast(eq_targets, 0, comms.world_comm)
            MPI.bcast(prob.iteration, 0, comms.world_comm)
            @debug "Rank $(comms.ranks) in iteration $(prob.iteration) calling the
            $(derivative_type_string(wrapper)) on wrapper $(equilibrium_type_string(wrapper))
            for state variables $var_indices"
            objective_gradient, _= MPI.recv(1, 1, comms.group_comm)
            target_gradients, _ = MPI.recv(1, 2, comms.group_comm)
            @debug "World head has received objective_gradient $objective_gradient
                    and target_gradients $target_gradients"
            for (i, v) in enumerate(var_indices)
                G[v] = objective_gradient[i]
                for (k, t) in target_gradients
                    prob.targets[k].gradient[v] = t[i]
                end
            end
            #foreach(p -> prob.targets[p.first].gradient = p.second, target_gradients)
            #foreach(p -> G[p[2]] = objective_gradient[p[1]], enumerate(var_indices))
        end
        MPI.bcast(true, 0, comms.world_comm)
    else
        while !finished_loop
            finished_loop = MPI.bcast(nothing, 0, comms.world_comm)
            if !finished_loop

                wrapper = MPI.bcast(nothing, 0, comms.world_comm)
                var_indices = MPI.bcast(nothing, 0, comms.world_comm)
                eq_targets = MPI.bcast(nothing, 0, comms.world_comm)
                iteration = MPI.bcast(nothing, 0, comms.world_comm)
                @debug "Rank $(comms.ranks) in iteration $iteration calling the
                        $(derivative_type_string(wrapper)) on wrapper $(equilibrium_type_string(wrapper))
                        for state variables $var_indices"
                if compute_rank == 0
                    objective_gradient, target_gradients =
                        gradient(deepcopy(wrapper), deepcopy(prob.state_vector[var_indices]),
                                 deepcopy(eq_targets), comms.compute_comm; iteration = iteration)
                    MPI.send(objective_gradient, 0, 1, comms.group_comm)
                    MPI.send(target_gradients, 0, 2, comms.group_comm)
                else
                    gradient(deepcopy(wrapper), deepcopy(prob.state_vector[var_indices]),
                    deepcopy(eq_targets), comms.compute_comm; iteration = iteration)
                end
            end
        end
    end
    return G
end

"""
    gradient(x::AbstractVector,prob::Problem,comms::OptComms)

The main function for computing the objective function and gradient value using MPI.
The global head node acts as a controller, sending worker groups which gradient index
to evaluate and gathering the results.  Only the global head node returns value of the
objective and gradient to the optimization algorithm.
"""
function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comms::OptComms;
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: ForwardDiff}
    world_rank = MPI.Comm_rank(comms.world_comm)
    group_rank = MPI.Comm_rank(comms.group_comm)
    compute_rank = MPI.Comm_rank(comms.compute_comm)
    n_worker_groups = MPI.Comm_size(comms.group_comm) - 1
    gradient_indices = world_rank == 0 ? collect(1:length(prob.state_vector)) : nothing
    # This is a potential vector for bugs if the tags collide due to a hard coded base tag
    recv_result_from_worker_tags = world_rank == 0 ? 19293 .+ collect(1:n_worker_groups) : nothing
    recv_targets_from_worker_tags = world_rank == 0 ? 14293 .+ collect(1:n_worker_groups) : nothing
    recv_index_from_head_tag = world_rank != 0 ? compute_rank == 0 ? 18323 + group_rank : nothing : nothing
    send_result_to_head_tag = world_rank != 0 ? compute_rank == 0 ? 19293 + group_rank : nothing : nothing
    send_targets_to_head_tag = world_rank != 0 ? compute_rank == 0 ? 14293 + group_rank : nothing : nothing
    send_index_to_worker_tags = world_rank == 0 ? 18323 .+ collect(1:n_worker_groups) : nothing
    recv_index_from_head_req = world_rank != 0 ? compute_rank == 0 ? MPI.Request() : nothing : nothing
    recv_result_from_worker_reqs = world_rank == 0 ? fill(MPI.Request(), n_worker_groups) : nothing
    index = zero(Int)
    index_buffer = zeros(T, 1)
    result = zero(T)
    result_buffer = zeros(T, 1)
    target_gradients = world_rank == 0 ? Dict{Int, Vector{T}}() : nothing
    target_gradient_buffer = world_rank == 0 ? Dict{Int, Vector{T}} : nothing
    foreach(k -> target_gradients[k] = zeros(T, length(state_variables)), keys(targets))
    index_map = world_rank == 0 ? Dict{Int, Int}() : nothing
    G = zeros(T, length(state_vector))

    # Do the gradient evaluation using the new base equilibrium
    # input parameters as the starting point
    # bcast to all processes
    finished_eval = MPI.bcast(false, 0, comms.world_comm)

    if world_rank != 0
        # This is a worker, doing the calculations
        index = 0
        if compute_rank == 0
            global recv_index_from_head_req, finished_eval
            while !finished_eval
                if MPI.isnull(recv_index_from_head_req)
                    # Initiate new request if current request is null
                    # The tag needs to match the sending tag from the global head
                    global recv_index_from_head_req = 
                        MPI.Irecv!(index_buffer, 0, recv_index_from_head_tag, comms.group_comm)
                    @debug "Initialize recv_from_head_req"

                else
                    # Valid request exists
                    recv_flag, _ = MPI.Test!(recv_index_from_head_req)
                    if recv_flag
                        global index = convert(Int, index_buffer[1])
                        MPI.bcast(index, 0, comms.compute_comm)
                        @debug "ComputeRank $compute_rank sent $index to workers"
                        if index >= 1
                            # Do the calculation
                            var = state_variables[index]
                            objective_derivative, target_derivatives = 
                                gradient(equilibrium_wrapper, deepcopy(var), targets, comms.compute_comm)
                            result_buffer[1] = objective_derivative
                            # Send the results back to the global head
                            MPI.Isend(result_buffer, 0, send_result_to_head_tag, comms.group_comm)
                            # Note this is a blocking send
                            MPI.Send(target_derivatives, 0, send_targets_to_head_tag, comms.group_comm)
                            # Reset the recvReq to a null Request
                            global recv_index_from_head_req = MPI.Request()
                        else
                            global finished_eval = true
                        end
                        MPI.bcast(finished_eval, 0, comms.compute_comm)
                    end
                end
            end
        else
            while !finished_eval
                logical_buffer = false
                index = MPI.bcast(nothing, 0, comms.compute_comm)
                if index >= 1
                    # Do the calculation
                    # index gets passed here, but instead split this up so that you can Mpi_send prob.equilibrium
                    var = state_variables[index]
                    _, _ = gradient(equilibrium_wrapper, deepcopy(var), targets, comms.compute_comm)
                end
                global finished_eval = MPI.bcast(logical_buffer, 0, comms.compute_comm)
            end
        end
        return nothing
    else
        foreach(k -> target_gradients[k] = zeros(T, length(state_variables)), keys(targets))
        while !finished_eval
            for group_index = 1:n_worker_groups
                # Start listening for receives
                if MPI.isnull(recv_results_from_worker_reqs[group_index])
                    grad_index = findfirst(x -> x != -1, gradient_indices)
                    if !isnothing(grad_index)
                        # Send out the gradient index to evaluate
                        @info "Evaluating gradient no. $(grad_index) out of $(length(prob.state_vector))"
                        index_buffer[1] = gradient_indices[grad_index]
                        index_map[group_index] = gradient_indices[grad_index]
                        gradient_indices[grad_index] = -1
                        # Initiate receive
                        recv_result_from_worker_reqs[group_index] = 
                            MPI.Irecv!(result_buffer, group_index,
                                       recv_result_from_worker_tags[group_index], 
                                       comms.group_comm)
                    end
                else
                    
                    # Valid request exists
                    recv_flag, _ = MPI.Test!(recv_result_from_worker_reqs[group_index])
                    if recv_flag
                        grad_index = index_map[group_index]
                        # Assign the gradient result if index > 0
                        result = result_buffer[1]                        
                        G[grad_index] = result
                        @debug "Recieved $(result) from  $group_index"
                        MPI.Recv!(target_gradient_buffer, group_index,
                                  recv_targets_from_worker_tags[group_index], comms.group_comm)
                        for key in keys(target_gradient_buffer)
                            target_gradients[key][grad_index] = target_gradient_buffer[key]
                        end
                        # If there are no more gradient points to compute, findfirst will return nothing
                        grad_index = findfirst(x -> x != -1, gradient_indices)
                        @debug gradient_indices
                        if !isnothing(grad_index)
                            @info "Evaluating gradient no. $(grad_index) out of $(length(prob.state_vector))"
                            index_buffer[1] = gradient_indices[grad_index]
                            index_map[group_index] = gradient_indices[grad_index]
                            gradient_indices[grad_index] = -1
                            # Initiate next receive
                            recv_result_from_worker_reqs[group_index] =
                                MPI.Irecv!(result_buffer, group_index,
                                           recv_result_from_worker_tags[group_index],
                                           comms.group_comm)
                        else
                            # Send the stop command to the workers
                            index_buffer[1] = -1
                            index_map[group_index] = -1
                            # No more work to do
                            recv_result_from_worker_reqs[group_index] = MPI.Request()
                        end
                        @debug "WorldRank $world_rank sending $(grad_index)  to group_index" group_index
                        MPI.Isend(index_buffer, group_index,
                                  send_index_to_worker_tags[group_index], comms.group_comm)
                    end
                end
            end
            if all(MPI.isnull.(recv_result_from_worker_reqs))
                global finished_eval = true
            end
        end
        return G, target_gradients
    end
end


function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_var::OptimizationVariable,
                  targets::Dict{Int, Target},
                  comm::MPI.Comm;
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: ForwardDiff}
    rank = MPI.Comm_rank(comm)
    g = equilibrium_wrapper.grad_method
    initial_value = copy(state_var.value)
    wrapper_input = deepcopy(equilibrium_wrapper.input)
    current_values = rank == 0 ? Dict{Int, T}() : nothing
    target_values = rank == 0 ? empty(current_values) : nothing
    target_gradients = rank == 0 ? empty(current_values) : nothing
    objective_gradient = rank == 0 ? zero(T) : nothing
    state_var.value = initial_value + g.δ
    @debug "$state_var"
    translate!(equilibrium_wrapper, [state_var])
    @debug "$(equilibrium_wrapper.input)"
    eq = compute_equilibrium(equilibrium_wrapper, comm.local_comm)
    
    if !isnothing(eq)
        update_equilibrium!(equilibrium_wrapper, eq)
        if rank == 0
            foreach(p -> current_values[p.first] =  p.second.current_value, targets)
            push!(target_values, compute_targets(targets, equilibrium_wrapper, comm)...)
            for key in keys(targets)
                target_gradients[key] = (target_values[key] - current_values[key]) / g.δ
                objective_gradient += ((targets[key].target_value - target_values[key])^2/targets[key].target_weight^2 -
                                  (targets[key].target_value - current_values[key])^2/targets[key].target_weight^2) / g.δ
            end
        else
            compute_targets(targets, equilibrium_wrapper, comm)
        end
    else
        if rank == 0
            foreach(k -> target_gradients[k] = convert(T, NaN), keys(targets))
            objective_gradient += NaN
        end
    end
    set_input!(equilibrium_wrapper, wrapper_input)

    return objective_gradient, target_gradients 

end

function gradient(equilibrium_wrapper::EquilibriumWrapper{T, E, D},
                  state_variables::Vector{OptimizationVariable},
                  targets::Dict{Int, Target},
                  comm::MPI.Comm;
                  iteration::Int = 1,
                  kwargs...
                 ) where {T, E <: AbstractMagneticEquilibrium, D <: SPSA}
    rank = MPI.Comm_rank(comm)
    g = equilibrium_wrapper.grad_method
    original_input = deepcopy(equilibrium_wrapper.input)
    target_keys = rank == 0 ? collect(keys(targets)) : nothing
    θ = map(v -> v.value, state_variables)
    δθ = similar(θ)
    target_gradients = rank == 0 ? Dict{Int, Vector{T}}() : nothing
    gₖ = rank == 0 ? zeros(T, length(θ)) : nothing
    gₖ_samples = rank == 0 ? zeros(T, (length(θ), g.n)) : nothing
    b_dist = rank == 0 ? Bernoulli() : nothing
    cₖ = rank == 0 ? g.δ/(iteration^g.γ) : nothing
    for n in 1:g.n
        valid_eq = false
        while !valid_eq
            if rank == 0
                forward_targets = Dict{Int, T}()
                backward_targets = Dict{Int, T}()
                Δ_targets = Dict{Int, Vector{T}}()
                Δ = rand(b_dist, length(state_variables)) |> x -> cₖ * (2x .- 1)
                δθ .= θ .+ Δ
                MPI.bcast(δθ, 0, comm)
                @debug "Rank $rank on iteration $n has forward perturbations $δθ"
                #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                for i in 1:length(δθ)
                    state_variables[i].value = δθ[i]
                    @debug "State variable $(name(state_variables[i])) on rank $rank for iteration $n has value $(state_variables[i].value)"
                end
                translate!(equilibrium_wrapper, state_variables)
                forward_eq = compute_equilibrium(equilibrium_wrapper, comm)
                if !isnothing(forward_eq)
                    update_equilibrium!(equilibrium_wrapper, forward_eq)
                    push!(forward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)

                    set_input!(equilibrium_wrapper, original_input)
                    δθ .= θ .- Δ
                    MPI.bcast(δθ, 0, comm)
                    @debug "Rank $rank on iteration $n has backward perturbations $δθ"

                    #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                    for i in 1:length(δθ)
                        state_variables[i].value = δθ[i]
                        @debug "State variable $(name(state_variables[i])) on rank $rank for iteration $n has value $(state_variables[i].value)"
                    end
                    translate!(equilibrium_wrapper, state_variables)
                    back_eq = compute_equilibrium(equilibrium_wrapper, comm)
                    if !isnothing(back_eq)
                        update_equilibrium!(equilibrium_wrapper, back_eq)
                        push!(backward_targets, compute_targets(targets, equilibrium_wrapper, comm)...)
                        for key in target_keys
                            forward_objective = (targets[key].target_value - forward_targets[key])^2/targets[key].target_weight^2
                            backward_objective = (targets[key].target_value - backward_targets[key])^2/targets[key].target_weight^2
                            Δ_targets[key] = 0.5 * (forward_targets[key] - backward_targets[key]) ./ Δ
                            gₖ_samples[:, n] .+= 0.5 * (forward_objective - backward_objective) ./ Δ
                        end
                        mergewith!(+, target_gradients, Δ_targets)
                        @debug "Δ_targets = $Δ_targets"
                        @debug "target_gradients: $target_gradients"
                        @debug "gₖ_samples[:, n] = $(gₖ_samples[:, n])"
                        valid_eq = true
                    end
                    set_input!(equilibrium_wrapper, original_input)
                else
                    set_input!(equilibrium_wrapper, original_input)
                end
                MPI.bcast(valid_eq, 0, comm)
            else
                δθ = MPI.bcast(nothing, 0, comm)
                @debug "Rank $rank on iteration $n has forward perturbations $δθ"
                #oreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                for i in 1:length(δθ)
                    state_variables[i].value = δθ[i]
                    @debug "State variable $(name(state_variables[i])) on rank $rank for iteration $n has value $(state_variables[i].value)"
                end
                translate!(equilibrium_wrapper, state_variables)
                forward_eq = compute_equilibrium(equilibrium_wrapper, comm)
                if !isnothing(forward_eq)
                    update_equilibrium!(equilibrium_wrapper, forward_eq)
                    compute_targets(targets, equilibrium_wrapper, comm)

                    set_input!(equilibrium_wrapper, original_input)
                    δθ = MPI.bcast(nothing, 0, comm)
                    @debug "Rank $rank on iteration $n has backward perturbations $δθ"

                    #foreach(p -> state_variables[p[1]].value = p[2], enumerate(δθ))
                    for i in 1:length(δθ)
                        state_variables[i].value = δθ[i]
                        @debug "State variable $(name(state_variables[i])) on rank $rank for iteration $n has value $(state_variables[i].value)"
                    end

                    translate!(equilibrium_wrapper, state_variables)
                    back_eq = compute_equilibrium(equilibrium_wrapper, comm)
                    if !isnothing(back_eq)
                        update_equilibrium!(equilibrium_wrapper, back_eq)
                        compute_targets(targets, equilibrium_wrapper, comm)
                    end
                    set_input!(equilibrium_wrapper, original_input)
                else
                    set_input!(equilibrium_wrapper, original_input)
                end
                valid_eq = MPI.bcast(nothing, 0, comm)
            end
        end
    end
    if rank == 0
        foreach(key -> target_gradients[key] .*= (1/g.n), keys(target_gradients))
        @debug "Final target_gradients: $target_gradients"
        aₖ = g.ϵ/((g.β + iteration)^g.α)
        gₖ[:] .= sum(gₖ_samples, dims = 2) ./ g.n
        @info "aₖ = $aₖ"
        @info "gₖ = $gₖ"
        @info "aₖ * gₖ = $(aₖ * gₖ)"
        return aₖ * gₖ, target_gradients
    end
end
